#======
# Copyright (c) IBM Corp. 2011, 2014.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


"""Updates the IBM PowerKVM system, given an update image or repository

This program should only be used, in case your system cannot directly access the Internet.
Otherwise, you should refer to the normal yum tools

Usage: ibm-update-system [OPTIONS] --iso-path=file
       ibm-update-system [OPTIONS] --url=repository

Options:

    -h
    --help
        Print this message and exit.

    -v
    --version
        Display version information and exit.

    -y
    --yes
        Assumes yes to answers

    -i
    --iso-path=
        Path to iso image containing updates

    -u
    --url=
        URL (http/ftp) to iso image containing updates


"""

import sys
import os
import getopt
import subprocess
import yum
import urllib
import urllib2
import shutil
from tempfile import mkdtemp
from yum import logging
from yum import logginglevels
from time import sleep
from model.installfunctions import upgradePackages
from yum.rpmtrans import RPMBaseCallback
from yum.constants import TS_INSTALL_STATES, TS_ERASE, TS_REMOVE_STATES
from contextlib import contextmanager

__version__ = "1.2"
__appname__ = "ibm-update-system"

#
# CONSTANTS AND DEFINITIONS
#
FOLLOW_PROGRESS = { "operation": "",
                    "percent": "",
                    "package": ""
                   }
__defproduct = "ibm_powerkvm"
__releasefile = '/etc/system-release'
__repofile = '/etc/yum/vars/ibmver'
__fprint_file = '.discinfo'
__gkey = 'file:///etc/pki/rpm-gpg/RPM-GPG-KEY-ibm_powerkvm'

def usage(code, msg=''):
    if code == 0:
        print >> sys.stderr, __doc__
    if msg:
        print >> sys.stderr, msg
    if code > 0:
        print >> sys.stderr, "Try '%s --help' for more information." % __appname__
    sys.exit(code)

def banner():
    sys.stdout.write("%s" % '='*int(len(__appname__) + len(__version__) + 1))
    sys.stdout.write("\n")
    sys.stdout.write("%s %s" % (__appname__, __version__))
    sys.stdout.write("\n")
    sys.stdout.write("%s" % '='*int(len(__appname__) + len(__version__) + 1))
    sys.stdout.write("\n")

def infoMsg(message):
    sys.stdout.write("%s\n" % message)

def getProductName():
    if os.path.isfile(__releasefile):
        with open(__releasefile, 'r') as f:
            line = f.readline()
            product = line.split('release')[0].strip().lower()
            msg = '\tProduct name: %s' % product
            infoMsg(msg)
            return product
    msg = '\tProduct name: %s' % __defproduct
    infoMsg(msg)
    return __defproduct

def getRepoVersion():
    if os.path.isfile(__repofile):
        with open(__repofile, 'r') as f:
            line = f.readline()  
            version = line.rstrip('\n').strip()
            msg = '\tProduct base version: %s' % version
            infoMsg(msg)
            return version
    return None

def checkVersions(iso_path):
    _filename = os.path.join(iso_path, __fprint_file)
    _msg = 'Invalid or inconsistent update image'
    if os.path.isfile(_filename):
        _pname = getProductName()
        _pversion = getRepoVersion()
        if _pversion != None:
            with open(_filename, 'r') as _file:
                for line in _file:
                    sline = line.lower().rstrip('\n').strip()
                    if sline.startswith(_pname):
                        updver = sline.split(_pname)[1].strip()
                        msg = '\tUpdate base version: %s' % updver
                        infoMsg(msg)
                        if updver.startswith(_pversion):
                            return
                _msg = 'Unknown or invalid update image version' 
        else:
            _msg = 'Could not identify system base version'
    raise RuntimeError(_msg)

def findRepo(mntpoint):
    repo = os.path.join(mntpoint, 'repodata')
    if os.path.exists(repo):
        return mntpoint
    else:
        repo = os.path.join(mntpoint, 'packages/repodata')
        if os.path.exists(repo):
            return os.path.join(mntpoint, 'packages')
    raise RuntimeError('Could not find a valid update repository')

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hvyi:u:',
                ['help', 'version', 'yes', 'iso-path=', 'url='])
    except getopt.error, msg:
        usage(1, msg)

    if not opts:
        usage(1, 'No options provided')

    if args:
        usage(1, 'Invalid extra argument(s) provided')

    repoURL = None
    assumeYes = False
    choosen = None

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-v', '--version'):
            print >> sys.stderr, __appname__ , __version__
            sys.exit(0)
        elif opt in ('-y', '--yes'):
            assumeYes = True
        elif opt in ('-i', '--iso-path'):
            if repoURL != None:
                usage(1, 'Should select only one update source (either iso or URL)')
            if not arg:
                usage(1, 'Missing path to update ISO image')
            repoURL = arg
            choosen = "iso"
        elif opt in ('-u', '--url'):
            if repoURL != None:
                usage(1, 'Should select only one update source (either iso or URL)')
            if not arg:
                usage(1, 'Missing URL to update image repository')
            repoURL = arg
            choosen = "url"
    if not repoURL:
        usage(1,'Missing path or URL to update image')

    banner()
    if not os.getuid() == 0:
        infoMsg("Must have root privileges")
        sys.exit(1)

    _tmpdir = setup()
    _xitcode = 0
    _mnt = False

    try:
        _mntpoint = mountISO(repoURL, _tmpdir, choosen)
        _mnt = True
        infoMsg("Checking versions...")
        checkVersions(_mntpoint)
        _repo = findRepo(_mntpoint)
        path = "file://" + _repo
        infoMsg("Checking available updates...")
        with pushd(_tmpdir):
            update(path, assumeYes, _tmpdir)
    except Exception, e:
        msg = 'Aborting... %s' % e
        infoMsg(msg)
        _xitcode = 1

    if _mnt:
        umountISO(_tmpdir)

    cleanup(_tmpdir)
    sys.exit(_xitcode)

@contextmanager
def pushd(dir):
    _cwd = os.getcwd()
    os.chdir(dir)
    yield
    os.chdir(_cwd)

def downloadISO(iso, tmpdir):
    file_name = os.path.join(tmpdir, iso.split('/')[-1])
    infoMsg("Downloading the image: " + iso)

    def dlProgress(count, blockSize, totalSize):
          percent = int(count*blockSize*100/totalSize)
          sys.stdout.write("\r" + iso + "...%d%%" % percent)
          sys.stdout.flush()

    request = urllib2.Request(iso)
    try:
        urllib2.urlopen(request)
    except urllib2.HTTPError as e:
        infoMsg('The specified server returned an error reponse.')
        raise
    except urllib2.URLError as e:
        infoMsg('Could not connect to the server.')
        raise

    urllib.urlretrieve (iso, file_name, reporthook=dlProgress)
    sys.stdout.write("\n")
    sys.stdout.flush()
    return file_name

def mountISO(iso, tmpdir, choosen):
    mount_point = os.path.join(tmpdir,'iso')
    os.makedirs(mount_point)

    if choosen == "url":
        iso = downloadISO(iso, tmpdir)

    infoMsg("Mounting image...")

    if not os.path.exists(iso):
        raise RuntimeError("File " + iso + " does not exist")

    cmd = "mount -o loop %s %s 2>/dev/null" % (iso, mount_point)

    if subprocess.call(cmd, shell=True) == 1:
        raise RuntimeError('Error mounting update image')

    return mount_point

def umountISO(tmpdir):

    mount_point = os.path.join(tmpdir,'iso')
    cmd = "umount %s 2>/dev/null" % mount_point

    if subprocess.call(cmd, shell=True) == 1:
        raise RuntimeError('Error umounting update image')

def setup():
    _tempwrk = mkdtemp('','ibmupd-','/tmp')
    return _tempwrk

def cleanup(tmpdir):

    shutil.rmtree(tmpdir)

def check_yes_or_no():
    input = None

    while input is None:
        input = raw_input("\nDo you want to proceed? (y/N)")

        if input.lower() == "y" or input.lower() == "yes":
            input = True
        elif input.lower() == "n" or input.lower() == "no":
            input = False
        else:
            input = None

    return input

def update(path, assumeYes, tmpdir):
    # instantiate yum object
    yb = yum.YumBase()

    # disable all repos
    yb.repos.disableRepo('*')

    basecache = os.path.join(tmpdir,'yum')
    os.makedirs(basecache)

    # configure yum parameters
    yb.conf.gpgcheck = True
    yb.conf.assumeyes = True
    yb.conf.cache = 0
    yb.conf.basecachedir = basecache

    # create a temporary repo to point to pkvm packages
    newrepo = yum.yumRepo.YumRepository('pkvm')
    newrepo.name = 'pkvm'
    newrepo.baseurl = [path]
    newrepo.basecachedir = yb.conf.basecachedir
    newrepo.gpgcheck = True
    newrepo.gpgkey = __gkey
    yb.repos.add(newrepo)
    yb.repos.enableRepo(newrepo.id)
    yb.doRepoSetup(thisrepo=newrepo.id)

    try:
        yb.pkgSack
    except RuntimeError as e:
        raise RuntimeError('Error creating yum instance: %s' % e)

    ylg = yum.logging.getLogger("yum.verbose.YumBase")
    ylg.setLevel(logging.CRITICAL)

    # mark all packages to install
    yb.repos.populateSack()
    yb.pkgSack.buildIndexes()

    updateObj = yb.up

    updatePackages = yb.doPackageLists('updates')

    update_package_list = []

    for i in updatePackages:
        update_package_list.append(i)

    update_package_list.sort()

    infoMsg('================================================================================')
    infoMsg('Packages')
    infoMsg('================================================================================')

    if update_package_list:
        print 'Updating:'
        for i in update_package_list:
            print '  ' + str(i)

        # mark packages to update
        for i in updatePackages:
            try:
                yb.update(i)
            except:
                pass


    obsolete_package_list = updateObj.getObsoletesList()
    # mark obsoleting packages to update
    if obsolete_package_list:
        print 'Obsoleting:'

        for pkg in obsolete_package_list:
            try:
                print '  ' + str(pkg[0])
                yb.update(name=pkg[0])
            except:
                pass

    # install new packages
    installed_pkgs = yb.doPackageLists().installed
    available_pkgs = yb.doPackageLists().available

    install_package_list = list(set(available_pkgs) - set(installed_pkgs) -\
            set(update_package_list)).sort()

    if install_package_list:
        print 'Installing'
        for i in install_package_list:
            print '  ' + str(i)

        if input:
            for pkg in install_package_list:
                try:
                    yb.install(pkg)
                except:
                    pass

    # execute
    try:
        if install_package_list or \
           update_package_list or \
           obsolete_package_list:

            print '================================================================================'
            if install_package_list:
               print 'Install: ' + str(len(install_package_list)) + ' pkgs'

            if update_package_list:
               print 'Update: ' + str(len(update_package_list)) + ' pkgs'

            if obsolete_package_list:
               print 'Obsoleting: ' + str(len(obsolete_package_list)) + ' pkgs'

            if not assumeYes:
                input = check_yes_or_no()
            else:
                input = True

            if input:
                yb.resolveDeps()
                yb.buildTransaction()
                yb.processTransaction(
                    None,
                    None,
                    UpdateCallback())
                print "\n"
        else:
            print 'No package updates found.'


    # cannot upgrade: re-raise exception
    except Exception, e:
        raise

    ylg.setLevel(logging.DEBUG)

    # close yum instance
    yb.closeRpmDB()
    yb.close()


class UpdateCallback(RPMBaseCallback):
    def __init__(self):
        """
        Constructor

        @type  callback: method
        @param callback: user callback
        """
        # call parent constructor
        RPMBaseCallback.__init__(self)

        # create start lock
        self.__startLock = {}
        self.__lastpackage = ""
    # __init__()

    def errorlog(self, msg):
        """
        Dummy method. Just ignores any error messages sent from the RPM
        installation process.

        @type  msg: basestring
        @param msg: RPM installation error messages

        @rtype: None
        @returns: nothing
        """
        pass
    # errorlog()

    def event(self, package, action, bytesProcessed, bytesTotal, currentProcess, totalProcess):
        """
        Handles installation progress

        @type  package: basestring
        @param package: A yum package object or simple string of a package name

        @type  action: int
        @param action: A yum.constant transaction set state or in the obscure
                      rpm repackage case it could be the string 'repackaging'

        @type  bytesProcessed: int
        @param bytesProcessed: Current number of bytes processed in the transaction
                           element being processed

        @type  bytesTotal: int
        @param bytesTotal: Total number of bytes in the transaction element being
                         processed

        @type  currentProcess: int
        @param currentProcess: number of processes completed in whole transaction

        @type  totalProcess: int
        @param totalProcess: total number of processes in the transaction.

        @rtype:   None
        @returns: Nothing
        """
        # action is install: get package name and save state
        if action in TS_INSTALL_STATES:
            packageSplited  = package._remote_url().split('/')
            packageName = packageSplited[ len(packageSplited) - 1]
            state = 'Processing'

        # state is uninstall: get package name and save state
        elif action in TS_REMOVE_STATES:
            packageName = package
            state = 'Cleanup'

        # other action: ignore
        else:
            return

        # calculate percentage
        percentage = (float(bytesProcessed) / float(bytesTotal) ) * 100

        # create message
        message = FOLLOW_PROGRESS

        # lock not exists: operation is starting
        if self.__startLock.get(packageName) == None:
            message['operation'] = '%s-start' % state
            self.__startLock[packageName] = True

        # percentage is 100: operation finish
        elif percentage == 100:
            message['operation'] = '%s-finish' % state

        # operation is running
        else:
            message['operation'] = '%s' % state

        # fill message dict
        message['percent'] = percentage
        message['package'] = packageName

        try:
            # break line when package change
            if self.__lastpackage != packageName:
                self.__lastpackage = packageName
                sys.stdout.write('\n')

            sys.stdout.write("\r%s (%s) %d%% [%d/%d]" % (state, packageName, percentage, currentProcess, totalProcess))
            sys.stdout.flush()

        except:
            print percentage
            pass

    # event()

    def filelog(self, package, action):
        """
        Dummy method. Just ignores any log messages sent from the RPM
        installation process.

        @type  package: basestring
        @param package: package name

        @type  action: basestring
        @param action: RPM installation log message

        @rtype: None
        @returns: nothing
        """
        pass
    # filelog()

    def scriptout(self, package, msgs):
        """
        Dummy method. Just ignores any scriptlet output messages sent from the
        RPM installation process.

        @type  package: basestring
        @param package: package name

        @type  msgs: basestring
        @param msgs: RPM scriptlet output messages

        @rtype: None
        @returns: nothing
        """
        pass
    # scriptout()

# UpdateCallback()

if __name__ == '__main__':
    main()
